/*
 * Author:  Lazar Lucian Ionut - 332CB
 * Project: Snake pe LCD cu bonusuri si nivele
 * File:    snake.c
 */

#include <avr/io.h>

#define F_CPU 16000000

#include <util/delay.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "u8g.h"

// game defines

#define LCD_HT 64
#define LCD_WH  128

#define UP    1
#define DOWN  2
#define LEFT  3
#define RIGHT 4

#define NONE  5
#define REVERSE_CONTROLS 7
#define REVERSE 8

#define INITIAL_SIZE 10
#define MAX_SIZE     300

#define SMALL_BLOCK_SIZE 2
#define LARGE_BLOCK_SIZE 4

#define MENU1CH1 1
#define MENU1CH2 2
#define MENU1CH3 3
#define MENU2CH1 4
#define MENU2CH2 5
#define MENU2CH3 6
#define START_GAME 7

uint8_t game_level = 1;

uint8_t menu;

uint8_t menu_direction;
uint8_t points = 0;

typedef struct {
    uint8_t x;
    uint8_t y;
} pos;

uint8_t snake_dir;

pos snake[MAX_SIZE];

pos food;

uint32_t head_pos;

uint32_t tail_pos;

uint32_t snake_size;

uint8_t bonus_points = 0;

uint8_t food_counter = 0;

uint8_t game_mode;

// LCD
u8g_t u8g;

int check_collisions(void) {
	int i;
	
	if (game_level >= 2) {
		for (i = 3; i < LCD_WH - 3; i++) {
			if (snake[head_pos].x == i && (snake[head_pos].y == 8 || snake[head_pos].y == 60)) {
				return 1;
			}
			
		}
	
		for (i = 8; i < LCD_HT - 4; i++) {
			if (snake[head_pos].y == i && (snake[head_pos].x == 5 || snake[head_pos].x == 123)) {
				return 1;
			}
		}
		
		if (game_level == 3) {		
			
			for (i = 20; i < 40; i++) {
				if (snake[head_pos].x == i && snake[head_pos].y == 20) {
					return 1;
				}
			}
			
			for (i = 80; i < 90; i++) {
				if (snake[head_pos].x == i && snake[head_pos].y == 50) {
					return 1;
				}
				
			}
			
			for (i = 70; i < 90; i++) {
				if (snake[head_pos].x == i && snake[head_pos].y == 18) {
					return 1;
				}
			}
			
			for (i = 40; i < 60; i++) {
				if (snake[head_pos].x == 15 && snake[head_pos].y == i) {
					return 1;
				}
			}
			
			for (i = 40; i < 60; i++) {
				if (snake[head_pos].x == 101 && snake[head_pos].y == i - 20) {
					return 1;
				}
			}
			
			for (i = 40; i < 54; i++) {
				if (snake[head_pos].x == 41 && snake[head_pos].y == i - 32) {
					return 1;
				}
			}	
		}
	}
	
	return 0;
}

void random_mode(void) {
	game_mode = rand() % 2 + 7;
	
	if(game_mode == REVERSE) {
		uint8_t i;
		for(i = 0; i < snake_size / 2; i++){
			uint32_t l = (tail_pos + i) % MAX_SIZE;
			uint32_t h = (head_pos - i) % MAX_SIZE;
			pos tmp;
			tmp = snake[l];
			snake[l] = snake[h];
			snake[h] = tmp;
		}
		
		if( snake_dir == UP){
			snake_dir = DOWN;
			return;
		}
		
		if( snake_dir == DOWN){
			snake_dir = UP;
			return;
		}
		
		if( snake_dir == LEFT){
			snake_dir = RIGHT;
			return;
		}
		
		if( snake_dir == RIGHT){
			snake_dir = LEFT;
			return;
		}
	}	
	
}

void init_buttons(void) {

    DDRA = 0xff;
    PORTA = 0x00;

    // UP
    DDRA &= ~(1 << PA0);
    PORTA |= (1 << PA0);
    
    // LEFT
    DDRA &= ~(1 << PA1);
    PORTA |= (1 << PA1);
    
    // RIGHT
    DDRA &= ~(1 << PA2);
    PORTA |= (1 << PA2);
    
    // DOWN
    DDRA &= ~(1 << PA3);
    PORTA |= (1 << PA3);
    
    // LED user
    DDRB |= (1 << PB7);
}

void reverse_control(void) {
	if (!(PINA & (1 << PA0))) {
		if( snake_dir == UP){
			return;
		}
        snake_dir = DOWN;
		return;
    }
    
    // LEFT
    if (!(PINA & (1 << PA1))) {
		if( snake_dir == LEFT){
			return;
		}
        snake_dir = RIGHT;
		return;
    }
    
    // RIGHT
    if (!(PINA & (1 << PA2))) {
		if( snake_dir == RIGHT){
			return;
		}
        snake_dir = LEFT;
		return;
    }
    
    // DOWN
    if (!(PINA & (1 << PA3))) {
		if( snake_dir == DOWN){
			return;
		}
        snake_dir = UP;
		return;
    }

}

/* Checks input. */
void check_input(void) {
	if( game_mode == REVERSE_CONTROLS){
		reverse_control();
		return;
	}
	
	menu_direction = -1;
    // UP
    if (!(PINA & (1 << PA0))) {
		if (menu == START_GAME) {
			if (snake_dir == DOWN)
				return;
			snake_dir = UP;
		}
		else
			menu_direction = UP;
		return;
    }
    
    // LEFT
    if (!(PINA & (1 << PA1))) {		
		if (menu == START_GAME) {
			if (snake_dir == RIGHT)
				return;
			snake_dir = LEFT;
		}
		else
			menu_direction = LEFT;
		return;
    }
    
    // RIGHT
    if (!(PINA & (1 << PA2))) {
		if (menu == START_GAME) {
			if (snake_dir == LEFT)
				return;
			snake_dir = RIGHT;
		}
		else
			menu_direction = RIGHT;
		return;
    }
    
    // DOWN
    if (!(PINA & (1 << PA3))) {	
		if (menu == START_GAME) {
			if (snake_dir == UP)
				return;
			snake_dir = DOWN;
		}
		else
			menu_direction = DOWN;
		return;
    }
	_delay_ms(10);
}

void init_lcd(void) {
    
    // init LCD - data ports PB0-PB7, control ports PC3-PC6 
    u8g_Init8Bit(&u8g, &u8g_dev_st7920_128x64_4x_8bit,
            PN(1, 0), PN(1, 1), PN(1, 2), PN(1, 3), 
            PN(1, 4), PN(1, 5), PN(1, 6), PN(1, 7), // DATA
            PN(2, 3),    // ENABLE
            U8G_PIN_NONE,
            U8G_PIN_NONE,
            PN(2, 5),    // RS / DI
            PN(2, 4),    // RW
            PN(2, 6));   // RESET
}

pos generate_pos(void) {
    pos c;
    
    c.x = rand() % LCD_WH;
    c.y = rand() % LCD_HT;
	
	/* Make sure food is between boundaries. */
	c.x = 7 + c.x % 50;
	c.y = 7 + c.y % 110;
    
    return c;
}

void init_snake(uint32_t seed) {
    uint8_t i;
    
	game_mode = NONE;
    snake_size = INITIAL_SIZE;
    
    for (i = 0; i < INITIAL_SIZE; i++) {
        snake[i].x = (LCD_WH - INITIAL_SIZE) / 2 + i * SMALL_BLOCK_SIZE;
        snake[i].y = LCD_HT / 2;
    }

    head_pos = INITIAL_SIZE - 1;
    tail_pos = 0;
	food_counter = 0;

    srand(seed);
    
    /* genereaza mancarea */
    food = generate_pos();
}

void update_game(void) {
    uint32_t prev;
 
    /* update pozitia capului */
    prev = head_pos;
    head_pos = (head_pos + 1) % MAX_SIZE;
    
    snake[head_pos] = snake[prev];
    
    switch(snake_dir) {
        case UP:
            snake[head_pos].y = (snake[head_pos].y + SMALL_BLOCK_SIZE) % LCD_HT;
            break;
        case DOWN:
            snake[head_pos].y = (snake[head_pos].y - SMALL_BLOCK_SIZE) % LCD_HT;
            break;
        case LEFT:
            snake[head_pos].x = (snake[head_pos].x - SMALL_BLOCK_SIZE) % LCD_WH;
            break;
        case RIGHT:
            snake[head_pos].x = (snake[head_pos].x + SMALL_BLOCK_SIZE) % LCD_WH;
            break;
    }
    
    /* verifica daca a mancat ceva */
    if (abs(food.x - snake[head_pos].x) < SMALL_BLOCK_SIZE && abs(food.y - snake[head_pos].y) < SMALL_BLOCK_SIZE) {
		if( food_counter == 0){
			game_mode = NONE;
		}
        snake_size++;
		points += 5;
		food_counter++;
		bonus_points++;
		
		
		if(food_counter == 4){
			food_counter = 0;
			random_mode();
		}
		if (bonus_points == 5) {
			points += 10;
			bonus_points = 0;
		}
        food = generate_pos();
    } else {
        tail_pos = (tail_pos + 1) % MAX_SIZE;
    }
}

/* Draw the initial menu. */
void draw_menu(void) {

	if (menu == MENU1CH1) {
		if (menu_direction == DOWN)
			menu = MENU1CH2;
		if (menu_direction == UP)
			menu = MENU1CH2;
		if (menu_direction == RIGHT) {
			menu = START_GAME;
			snake_dir = RIGHT;
		}
		goto next;
	}
	if (menu == MENU1CH2) {
		if (menu_direction == DOWN)
			menu = MENU1CH1;
		if (menu_direction == UP)
			menu = MENU1CH1;
		if (menu_direction == RIGHT)
			menu = MENU2CH1;		
		goto next;
	}
		
	if(menu == MENU2CH1) {
		if (menu_direction == DOWN)
			menu = MENU2CH3;
		if (menu_direction == UP)
			menu = MENU2CH2;
		if (menu_direction == LEFT) {
			game_level = 1;				
			menu = MENU1CH2;
		}
		goto next;
	}
	if(menu == MENU2CH2) {
		if (menu_direction == DOWN)
			menu = MENU2CH1;
		if (menu_direction == UP)
			menu = MENU2CH3;
		if (menu_direction == LEFT) {
			game_level = 2;					
			menu = MENU1CH2;
		}
		goto next;
	}	
	if(menu == MENU2CH3) {
		if (menu_direction == DOWN)
			menu = MENU2CH2;
		if (menu_direction == UP)
			menu = MENU2CH1;
		if (menu_direction == LEFT) {
			game_level = 3;
			menu = MENU1CH2;
		}
		goto next;
	}	
next:

	menu_direction = -1;
	
	switch (menu) {
		case MENU1CH1:
			u8g_SetDefaultForegroundColor(&u8g);
			u8g_SetFont(&u8g, u8g_font_helvB10);
			u8g_DrawStr(&u8g, 5, 15, ">Start joc");				
			u8g_DrawStr(&u8g, 5, 30, "Alege nivel");	
			break;
		case MENU1CH2:
			u8g_SetDefaultForegroundColor(&u8g);
			u8g_SetFont(&u8g, u8g_font_helvB10);
			u8g_DrawStr(&u8g, 5, 15, "Start joc");				
			u8g_DrawStr(&u8g, 5, 30, ">Alege nivel");	
			break;
			break;
		case MENU2CH1:
			u8g_SetDefaultForegroundColor(&u8g);
			u8g_SetFont(&u8g, u8g_font_helvB10);
			u8g_DrawStr(&u8g, 5, 15, ">Usor");				
			u8g_DrawStr(&u8g, 5, 30, "Mediu");				
			u8g_DrawStr(&u8g, 5, 45, "Dificil");
			break;

		case MENU2CH2:
			u8g_SetDefaultForegroundColor(&u8g);
			u8g_SetFont(&u8g, u8g_font_helvB10);
			u8g_DrawStr(&u8g, 5, 15, "Usor");				
			u8g_DrawStr(&u8g, 5, 30, ">Mediu");				
			u8g_DrawStr(&u8g, 5, 45, "Dificil");
			break;
		case MENU2CH3:
			u8g_SetDefaultForegroundColor(&u8g);
			u8g_SetFont(&u8g, u8g_font_helvB10);
			u8g_DrawStr(&u8g, 5, 15, "Usor");				
			u8g_DrawStr(&u8g, 5, 30, "Mediu");				
			u8g_DrawStr(&u8g, 5, 45, ">Dificil");
			break;
		default:
			break;	
	}
    
}

/* End game screen. */
void draw_end_game_screen(void) {
	char *str = malloc(15);

	u8g_SetDefaultForegroundColor(&u8g);
    u8g_SetFont(&u8g, u8g_font_helvB10);
    u8g_DrawStr(&u8g, 40, 20, "Ai pierdut!");
	sprintf(str, "%d", points);
	strcat(str, " puncte");
	u8g_DrawStr(&u8g, 50, 50, str);
	
	free(str);

}

void draw_screen(void) {
    uint32_t i;
    pos block;
    

    u8g_SetDefaultForegroundColor(&u8g);
	
	/* Draw the snake. */
	
	for (i = 0; i < snake_size; i++) {
		block = snake[(tail_pos + i) % MAX_SIZE];
		u8g_DrawBox(&u8g, block.x - SMALL_BLOCK_SIZE / 2, block.y - SMALL_BLOCK_SIZE / 2, SMALL_BLOCK_SIZE, SMALL_BLOCK_SIZE);		
	}
    
    /* Draw the food block. */
    u8g_DrawBox(&u8g, food.x - SMALL_BLOCK_SIZE / 2, food.y - SMALL_BLOCK_SIZE / 2, SMALL_BLOCK_SIZE, SMALL_BLOCK_SIZE);
	
	
	
	/* Level */
	if (game_level >= 2) {
			/* Draw boundary. */
		for (i = 0; i < LCD_WH; i++) {
			u8g_DrawBox(&u8g, i, 5, SMALL_BLOCK_SIZE / 2, SMALL_BLOCK_SIZE / 2);
			u8g_DrawBox(&u8g, i, 63, SMALL_BLOCK_SIZE / 2, SMALL_BLOCK_SIZE / 2);
		}
		
		for (i = 5; i < LCD_HT; i++) {
			u8g_DrawBox(&u8g, 0, i, SMALL_BLOCK_SIZE / 2, SMALL_BLOCK_SIZE / 2);
			u8g_DrawBox(&u8g, 127, i, SMALL_BLOCK_SIZE / 2, SMALL_BLOCK_SIZE / 2);
		}
		/* Draw external boundary. */
		for (i = 2; i < LCD_WH - 3; i++) {
			u8g_DrawBox(&u8g, i, 7, SMALL_BLOCK_SIZE, SMALL_BLOCK_SIZE);
			u8g_DrawBox(&u8g, i, 60, SMALL_BLOCK_SIZE, SMALL_BLOCK_SIZE);
		}
	
		for (i = 8; i < LCD_HT - 3; i++) {
			u8g_DrawBox(&u8g, 2, i, SMALL_BLOCK_SIZE, SMALL_BLOCK_SIZE);
			u8g_DrawBox(&u8g, 124, i, SMALL_BLOCK_SIZE, SMALL_BLOCK_SIZE);
		}
		
		if (game_level == 3) {
			/* Draw the level 3 barriers */
			
			for (i = 20; i < 40; i++) {
				u8g_DrawBox(&u8g, i, 20, SMALL_BLOCK_SIZE / 2, SMALL_BLOCK_SIZE / 2);
			}
			
			for (i = 80; i < 90; i++) {
				u8g_DrawBox(&u8g, i, 50, SMALL_BLOCK_SIZE / 2, SMALL_BLOCK_SIZE / 2);
			}
			
			for (i = 70; i < 90; i++) {
				u8g_DrawBox(&u8g, i, 18, SMALL_BLOCK_SIZE / 2, SMALL_BLOCK_SIZE / 2);
			}
			
			for (i = 40; i < 60; i++) {
				u8g_DrawBox(&u8g, 14, i, SMALL_BLOCK_SIZE / 2, SMALL_BLOCK_SIZE / 2);
			}
			
			for (i = 40; i < 60; i++) {
				u8g_DrawBox(&u8g, 100, i - 20, SMALL_BLOCK_SIZE / 2, SMALL_BLOCK_SIZE / 2);
			}
			
			for (i = 40; i < 54; i++) {
				u8g_DrawBox(&u8g, 40, i - 32, SMALL_BLOCK_SIZE / 2, SMALL_BLOCK_SIZE / 2);
			}	
		}
	
	}
}

int main(void) {
	uint32_t i, check_die, j;
    init_buttons();
    init_lcd();
	
	menu = 1;
	menu_direction = 0;
    
    snake_dir = NONE;
    
    /* Menu loop */
    while(1) {
        check_input();
        
        if (menu == START_GAME) break;
        
        u8g_FirstPage(&u8g);
        do {
            draw_menu();
        } while (u8g_NextPage(&u8g));        
        
		_delay_ms(10);
    }
    
    init_snake(14);
	snake_dir = RIGHT;

	/* Game Loop */
    while (1) {
        check_input();
        update_game();
        
        u8g_FirstPage(&u8g);
        do {
            draw_screen();
        } while (u8g_NextPage(&u8g));
		
		/* Check if the snake eats itself. */
		check_die = 0;
		for(i = 1; i < snake_size; i++){
			j = (head_pos - i) % MAX_SIZE;
			if(abs(snake[j].x - snake[head_pos].x) < SMALL_BLOCK_SIZE && abs(snake[j].y - snake[head_pos].y) < SMALL_BLOCK_SIZE){
				check_die++;
				break;
			}
		}
		
		/* Check if the snake hits any boundaries. */		
		check_die += check_collisions();
		
		/* If so show end game screen. */
		if (check_die > 0){
			break;
		}
    }
	
	/* Draw end screen */
	while(1) {
        u8g_FirstPage(&u8g);
        do {
            draw_end_game_screen();
        } while (u8g_NextPage(&u8g));    
        
        _delay_ms(10);
    }
    
    return 0;
}
